import json
import os
import typer
from openai import OpenAI
import agent.tools as tools
from agent.logging import logger
from agent.rich import print_in_panel, print_in_question_panel
from agent.utils import check_for_screenshots
from agent.config import SUPPORTED_LANGUAGES, get_test_dir


def language_prompt(language: str):
    if language.lower() not in SUPPORTED_LANGUAGES:
        logger.error(f"Language {language} not supported.")
        raise typer.Exit()
    return f"You should write tests using the {language} programming language."


def framework_prompt(framework: str, language: str = "python"):
    if framework == "playwright" and language == "python":
        return """Tests should be written using playwright. We are using the pytest-playwright plugin, so you don't need to create your own browser context. Use page.locator where possible - use classnames, ids, placeholders, visible text or even labels to find elements. Avoid using deprecated methods.

        The structure of each test should be as follows:
        
        ```
        from playwright.sync_api import Page

        def test_name(page: Page):
            # test code here
        
        ```
        """
    elif framework == "playwright":
        return """
        Tests should be written using playwright. Use page.locator where possible - use classnames, ids, placeholders, visible text or even labels to find elements. Avoid using deprecated methods. Avoid imports from any external libraries unless the user specifies it.

        The structure of each test should be as follows:
        ```
        import { test, expect } from '@playwright/test';

        test('has title', async ({ page }) => {
            await page.goto('https://playwright.dev/');

            // Expect a title "to contain" a substring.
            await expect(page).toHaveTitle(/Playwright/);
        });
        ```
        """
    elif framework == "cypress":
        return """
        Tests should be written using cypress. Avoid using deprecated methods. Avoid imports from any external libraries unless the user specifies it.

        The structure of each test should be as follows:
        ```
        describe('My First Test', () => {
            it('Gets, types and asserts', () => {
                cy.visit('https://example.cypress.io')

                cy.contains('type').click()

                // Should be on a new URL which
                // includes '/commands/actions'
                cy.url().should('include', '/commands/actions')

                // Get an input, type into it
                cy.get('.action-email').type('fake@email.com')

                //  Verify that the value has been updated
                cy.get('.action-email').should('have.value', 'fake@email.com')
            })
        })
        ```
        """
    else:
        return ""


def log_completion(res: dict):
    # try to get ai message content
    try:
        content = json.loads(res)["choices"][0]["message"]["content"]
    except KeyError:
        content = None

    # try to get tool function calls

    try:
        tool_calls = json.loads(res)["choices"][0]["message"]["tool_calls"]
    except KeyError:
        tool_calls = None

    # log info in a panel
    if content:
        print_in_panel(str(content), title="Agent Message")

    if tool_calls:
        # loop through tool calls and format them as python function calls
        for tool_call in tool_calls:
            name = tool_call["function"]["name"]
            kwargs = json.loads(tool_call["function"]["arguments"])
            # truncate any long arguments
            for key, value in kwargs.items():
                if len(str(value)) > 50:
                    kwargs[key] = f"{str(value)[:50]}..."

            print_in_panel(f"{name}({kwargs})", title="Tool Call Requested")


def create_completion(messages: list, client: OpenAI):
    response = client.chat.completions.create(
        model=os.environ.get("OPENAI_MODEL", "gpt-4o"),
        messages=messages,
        tools=tools.tools,
        tool_choice="auto",
        temperature=0.3,
    )
    log_completion(response.to_json())
    tool_calls = response.choices[0].message.tool_calls
    if tool_calls is None:
        tool_calls = []

    return response, tool_calls


def agent(
    prompt: str, url: str, language: str = "python", framework: str = "playwright"
):
    client = None
    openai_api_key = os.environ.get("OPENAI_API_KEY")

    if openai_api_key is None:
        return logger.error(
            "No OpenAI API key found. Please set the [purple3]OPENAI_API_KEY[/purple3] environment variable."
        )

    client = OpenAI(api_key=openai_api_key)

    agent_working = True
    messages = []
    messages.append(
        {
            "role": "assistant",
            "content": "You are a useful test code writing agent. Use the supplied tools to create passing tests for the user.",
        },
    )
    messages.append(
        {
            "role": "assistant",
            "content": f"{language_prompt(language)} {framework_prompt(framework)}",
        }
    )
    messages.append(
        {
            "role": "assistant",
            "content": f"""If tests are not passing, consider using the tools to debug the issue. 
            - You can overwrite any code in the {get_test_dir()} folder using the relevant tools. Use links found on the webpage to determine if navigation to other pages are required. 
            - You can use the extract_webpage_content tool in place of navigation. 
            - Do not make assumptions about the app structure or redirects unless there are clear links to support it. If you need more context, add code to save screenshots to the 'test-results' folder and the run the tests. We will send you the screenshots.
            - If you need input from the user, always use the get_user_input tool.""",
        },
    )
    messages.append(
        {
            "role": "user",
            "content": f"Our webpage entry point is: {url}. Consider the following requirements: {prompt}",
        },
    )
    messages.append(
        {
            "role": "user",
            "content": f"Consider the following requirements: {prompt}",
        },
    )

    logger.info(
        "Starting agent. Set log level to [cadet_blue]DEBUG[/cadet_blue] to see full requests."
    )
    logger.debug(json.dumps(messages, indent=2))

    response, tool_calls = create_completion(messages, client)

    while agent_working:
        if tool_calls == []:
            logger.info("No obvious action to be taken.")
            # prompt the user for an additional prompt
            print_in_question_panel(
                """Please enter an additional prompt if you would like to give the agent more context. If you have no additional prompt, you can type exit to stop the agent.""",
                title="Additional Prompt",
            )
            user_input = typer.prompt("Additional Prompt")
            if user_input.lower() != "exit":
                messages.append(
                    {
                        "role": "user",
                        "content": user_input,
                    }
                )
                response, tool_calls = create_completion(messages, client)
            else:
                agent_working = False
                break

        for tool_call in tool_calls:
            name = tool_call.function.name
            kwargs = json.loads(tool_call.function.arguments)
            logger.debug(f"Calling tool: {name} with arguments: {kwargs}")

            function_call_output = getattr(tools, name)(**kwargs)
            messages.append(
                {
                    "role": "assistant",
                    "content": f"Called tool: {name} with arguments: {kwargs}. The function call output was: {function_call_output} ",
                }
            )
            if name == "run_tests":
                screenshots = check_for_screenshots()
                for screenshot in screenshots:
                    logger.info("Screenshot found. Adding to context.")
                    messages.append(
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "text",
                                    "text": "Screenshots were taken on test failure. Please review the screenshots below to help debug the failing tests.",
                                },
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": f"data:image/jpeg;base64,{screenshot}"
                                    },
                                },
                            ],
                        }
                    )

            logger.debug(json.dumps(messages, indent=2))
            response, tool_calls = create_completion(messages, client)



import pytest
from pathlib import Path
from agent.scaffold import scaffold_cypress, delete_entire_dir


@pytest.mark.parametrize("language", ["typescript", "javascript"])
def test_cypress_scaffolding(language: str):
    test_dir = Path(f"test_cypress_{language}")

    scaffold_cypress(
        test_dir=test_dir,
        language=language,
        clean=True,
    )

    files_in_test_dir = list(test_dir.glob("*"))

    node_modules = test_dir / "node_modules"
    e2e_folder = test_dir / "e2e"
    support_folder = test_dir / "support"

    # check for the node_modules folder
    assert node_modules.exists()
    # check for an e2e folder
    assert e2e_folder.exists()
    # check for support folder
    assert support_folder.exists()

    # check for relevant files
    assert any("package.json" in str(f) for f in files_in_test_dir)
    assert any(".gitignore" in str(f) for f in files_in_test_dir)

    if language == "javascript":
        assert any("cypress.config.js" in str(f) for f in files_in_test_dir)
        # check for the support/commands.js file
        commands_file = support_folder / "commands.js"
        assert commands_file.exists()

        # check for the support/e2e.js file
        e2e_file = support_folder / "e2e.js"
        assert e2e_file.exists()

    if language == "typescript":
        assert any("tsconfig.json" in str(f) for f in files_in_test_dir)
        assert any("cypress.config.ts" in str(f) for f in files_in_test_dir)

        # check for the support/commands.js file
        commands_file = support_folder / "commands.ts"
        assert commands_file.exists()

        # check for the support/e2e.js file
        e2e_file = support_folder / "e2e.ts"
        assert e2e_file.exists()

    # delete the test directory
    delete_entire_dir(test_dir)



import os
import typer
from playwright.sync_api import sync_playwright
from agent.config import get_test_dir
from agent.rich import print_in_panel, print_in_question_panel
from agent.utils import (
    run_cypress,
    run_playwright,
    run_pytest_playwright,
    strip_code_fences,
)
from agent.logging import logger
from typing import TypedDict
from bs4 import BeautifulSoup


# TODO: Consider adding pydantic for additional data validation
class TExtractWebpageContent(TypedDict):
    url: str


def extract_webpage_content(**kwargs: TExtractWebpageContent) -> str | None:
    url = kwargs.get("url", None)
    if not url:
        return logger.error("No URL provided to extract webpage content from.")

    logger.info(f"Extracting webpage content from {url}")
    content = None
    with sync_playwright() as p:
        logger.info("Launching browser")
        browser = p.chromium.launch(
            headless=os.environ.get("HEADLESS", False) == "true"
        )
        page = browser.new_page()
        logger.info(f"Visiting {url}")
        page.goto(url)
        logger.info("Waiting for page load state to be domcontentloaded")
        page.wait_for_load_state("domcontentloaded")
        logger.info("Page loaded! Extracting content.")
        content = page.content()

        # use beautiful soup to extract the body only
        soup = BeautifulSoup(content, "html.parser")
        # delete the head tag
        soup.head.extract()
        # delete the script tags
        for script in soup.find_all("script"):
            script.extract()

        content = str(soup)

        page_file_name = f"{url.replace('https://', '').replace('http://', '').replace('/', '_')}.html"

        if os.environ.get("LOG_LEVEL", "INFO") == "DEBUG":
            logger.info(f"Writing content to {page_file_name}")
            test_dir = get_test_dir()
            with open(test_dir / page_file_name, "w+") as f:
                f.write(content)

        logger.info("Closing browser")
        browser.close()

    return content


class TWriteCodeToFile(TypedDict):
    code: str
    file_name: str


def write_code_to_file(**kwargs: TWriteCodeToFile):
    code = kwargs.get("code", None)
    file_name = kwargs.get("file_name", None)

    if not code or not file_name:
        error_message = "code or file_name not provided to write code to file."
        logger.error(error_message)
        return error_message

    # TODO: Make this more robust
    # get language from environment
    language = os.environ.get("AGENT_LANGUAGE", "python")
    # get framework from environment
    framework = os.environ.get("AGENT_FRAMEWORK", "playwright")

    invalid_characters = ["\\", "/", ":", "*", "#", "?", '"', "'" "<", ">", "|"]
    # check file_name for invalid characters

    for char in invalid_characters:
        if char in file_name:
            return f"Invalid file name. Please provide a valid file name without any of the following invalid characters. {invalid_characters}"

    file_extension = ".py"

    if language == "typescript":
        file_extension = ".ts"
        if framework == "cypress":
            file_extension = ".cy.ts"
        elif framework == "playwright":
            file_extension = ".spec.ts"
    elif language == "javascript":
        file_extension = ".js"
        if framework == "cypress":
            file_extension = ".cy.js"
        elif framework == "playwright":
            file_extension = ".spec.js"

    if not file_name.endswith(file_extension):
        return f"Invalid language or file extension provided. Please rewrite the code in {language} and use a {file_extension} extension."

    out_dir = get_test_dir()

    if framework == "cypress":
        out_dir = out_dir / "e2e"
        # ensure the tests directory exists
        out_dir.mkdir(exist_ok=True)
    elif framework == "playwright":
        # ensure the tests directory exists
        out_dir = out_dir / "tests"
        out_dir.mkdir(exist_ok=True)

    logger.info(f"Writing code to file {file_name}")
    with open(out_dir / file_name, "w+") as f:
        f.write(strip_code_fences(code))
        return f"Code written to {file_name}"


def run_tests():
    # ensure the tests directory exists
    test_dir = get_test_dir()
    test_dir.mkdir(exist_ok=True)
    logger.info("Running tests")

    # get the language and framework from the environment
    language = os.environ.get("AGENT_LANGUAGE", "python")
    framework = os.environ.get("AGENT_FRAMEWORK", "playwright")
    if language == "python" and framework == "playwright":
        # run the tests and capture the output
        output = run_pytest_playwright(test_dir)
        # print output in a panel
        print_in_panel(output, "Test Output")
        return output

    elif language == "typescript" and framework == "playwright":
        output = run_playwright(test_dir / "tests")
        print_in_panel(output, "Test Output")
        return output
    elif language == "javascript" and framework == "playwright":
        output = run_playwright(test_dir / "tests")
        print_in_panel(output, "Test Output")
        return output
    elif language == "typescript" and framework == "cypress":
        output = run_cypress(test_dir, config_file_name="cypress.config.ts")
        print_in_panel(output, "Test Output")
        return output
    elif language == "javascript" and framework == "cypress":
        output = run_cypress(test_dir, config_file_name="cypress.config.js")
        print_in_panel(output, "Test Output")
        return output
    else:
        return "Not possible to run tests."


class TGetUserInput(TypedDict):
    question: str


def get_user_input(**kwargs: TGetUserInput) -> str:
    question = kwargs.get("question", None)

    if not question:
        return logger.error("No question provided to get user input for.")

    print_in_question_panel(question)
    response = typer.prompt("Response")

    return response


tools = [
    {
        "type": "function",
        "function": {
            "name": "run_tests",
            "description": f"Runs the written tests in the '{get_test_dir()}' folder. Call this whenever you need to validate if the tests are working as expected.",
        },
    },
    {
        "type": "function",
        "function": {
            "name": "write_code_to_file",
            "description": f"Writes the generated code to a file. Once the tests are written to the file, you will be able to run them using 'run_tests'. The file will be created in the '{get_test_dir()}' folder. We are using the w+ mode to write the file.",
            "parameters": {
                "type": "object",
                "properties": {
                    "code": {
                        "type": "string",
                        "description": "The generated code to be written to a file. Should be plain text and not include code fences.",
                    },
                    "file_name": {
                        "type": "string",
                        "description": "The name of the file to write the code to. Including the file extension.",
                    },
                },
                "required": ["code", "file_name"],
                "additionalProperties": False,
            },
        },
    },
    {
        "type": "function",
        "function": {
            "name": "extract_webpage_content",
            "description": "Extracts the content of a webpage.",
            "parameters": {
                "type": "object",
                "properties": {
                    "url": {
                        "type": "string",
                        "description": "The URL of the webpage to extract content from.",
                    },
                },
                "required": ["url"],
                "additionalProperties": False,
            },
        },
    },
    {
        "type": "function",
        "function": {
            "name": "get_user_input",
            "description": "Prompts the user for input. Use this tool to get input from the user when needed.",
            "parameters": {
                "type": "object",
                "properties": {
                    "question": {
                        "type": "string",
                        "description": "The question to ask the user for input.",
                    },
                },
                "required": ["question"],
                "additionalProperties": False,
            },
        },
    },
]



import typer
from typing_extensions import Annotated
from typing import Optional
from pathlib import Path
from agent.config import get_test_dir
from agent.logging import console
from rich.panel import Panel
from rich.markdown import Markdown


def init(
    folder: Annotated[
        Optional[Path],
        typer.Argument(help="The folder to initialise the project in."),
    ] = ".",
):
    """Initialise a new directory for the project 🚀"""

    if str(folder) == ".":
        # ask for confirmation
        confirm = typer.confirm(
            "Do you want to initialise the project in the current directory?"
        )
        if not confirm:
            raise typer.Exit()

    # make the project directory if it doesn't exist
    folder.mkdir(exist_ok=True)

    # create the tests folder if it doesn't exist
    test_dir = get_test_dir()
    test_dir = folder / test_dir
    test_dir.mkdir(exist_ok=True)

    # create an .env file if it doesn't exist
    env_dir = folder / ".env"
    if not (env_dir).exists():
        with open(env_dir, "w") as f:
            f.write("OPENAI_API_KEY=''\n")

    # create a .gitignore file if it doesn't exist
    gitignore_dir = folder / ".gitignore"
    if not (gitignore_dir).exists():
        with open(gitignore_dir, "w") as f:
            f.write(".env\n")

    # create a config.toml file if it doesn't exist
    agent_config_dir = folder / "config.toml"
    if not (agent_config_dir).exists():
        with open(agent_config_dir, "w") as f:
            config_toml = r"""
[config]
language = "python"
framework = "playwright"
headless = false
clean = true
log_level = "INFO"

[agent]
url = "https://dev.datamaker.app"
prompt = '''
Test the signup and sign in flow of the app
'''
"""
            f.write(config_toml)
    console.print("\n")
    console.print(
        Panel(
            Markdown(
                f"""
You can start writing tests using the agent CLI. Run the following commands to get started:
- cd {folder}
- agent start

You may also need to add your *OPENAI_API_KEY* to the **.env** file. Modify your *prompt* in the **config.toml** file
"""
            ),
            padding=(1, 1),
            title="New project initialised successfully! 🚀",
            highlight=True,
            title_align="left",
        )
    )
    console.print("\n")



from agent.logging import console
from rich.panel import Panel


def print_in_question_panel(text: str, title: str = "Question?"):
    console.print("\n")
    console.print(
        Panel(
            text,
            title=title,
            padding=(1, 1),
            highlight=True,
            border_style="color(49)",
        )
    )
    console.print("\n")


def print_in_panel(text: str, title: str = "Output"):
    console.print("\n")
    console.print(
        Panel(
            text,
            title=title,
            padding=(1, 1),
            highlight=True,
        )
    )
    console.print("\n")



import typer
from rich import print
from agent.cli import init, start
from agent import __version__

app = typer.Typer(name="agent", rich_markup_mode="markdown")
app.command(short_help="Scaffold new project.")(init.init)
app.command(short_help="Start the agent.")(start.start)


def version_callback(value: bool):
    if value:
        print(
            f"Automators Agent CLI Version:  [medium_spring_green]{__version__}[/medium_spring_green]"
        )
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=version_callback),
):
    pass


if __name__ == "__main__":
    app()



import os
import importlib.metadata
from pathlib import Path
from dotenv import load_dotenv

__app_name__ = "agent"
__version__ = importlib.metadata.version("automators-agent")

dotenv_path = Path(os.getcwd()) / ".env"
load_dotenv(dotenv_path=dotenv_path, override=True)



import logging
import os
from rich.logging import RichHandler
from rich.console import Console
from rich.theme import Theme

console = Console(
    theme=Theme(
        {
            "logging.level.info": "medium_spring_green",
            "logging.level.debug": "cadet_blue",
            "logging.level.warning": "yellow1",
            "logging.level.error": "indian_red1",
        }
    ),
)

FORMAT = "%(message)s"
logging.basicConfig(
    level=os.environ.get("LOG_LEVEL", "INFO"),
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(markup=True, console=console, rich_tracebacks=True)],
)

logger = logging.getLogger("agent")



import os
import typer
from typing_extensions import Annotated
from rich.console import Console
from agent.completions import agent
from agent.logging import logger
from agent.config import get_test_dir, read_config
from agent.scaffold import (
    check_for_cypress_installation,
    check_for_node,
    check_for_npm,
    check_for_playwright,
    check_for_playwright_browsers,
    scaffold_playwright,
    scaffold_cypress,
)

err_console = Console(stderr=True)


def start(
    prompt: Annotated[
        str, typer.Option(help="A prompt describing the scope of your tests.")
    ] = None,
    url: Annotated[
        str,
        typer.Option(help="A url that acts as an entrypoint to app you're testing."),
    ] = None,
    clean: Annotated[
        bool,
        typer.Option(
            help="Delete files from the relevant test output directory before starting the agent."
        ),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option(help="Enable debug mode for additional logs and file writing."),
    ] = False,
    headless: Annotated[
        bool,
        typer.Option(help="Run the browser in headless mode."),
    ] = False,
    language: Annotated[
        str,
        typer.Option(help="The programming language to generate tests in."),
    ] = "python",
    framework: Annotated[
        str,
        typer.Option(help="The framework to use for testing."),
    ] = "playwright",
):
    """Starts the test generation agent ✨"""

    config = read_config()

    # CLI args should take precedence over config file
    if prompt is not None:
        config["agent"]["prompt"] = prompt
    if url is not None:
        config["agent"]["url"] = url
    if headless:
        config["config"]["headless"] = headless
    if debug:
        config["config"]["log_level"] = "DEBUG"
    if clean:
        config["config"]["clean"] = clean
    if language != "python":
        config["config"]["language"] = language
    if framework != "playwright":
        config["config"]["framework"] = framework

    # perform actions based on config
    log_level = str(config["config"]["log_level"]).upper()

    # set a defualt log level if invalid
    if log_level not in ["DEBUG", "INFO", "WARN", "ERROR"]:
        log_level = "INFO"

    logger.setLevel(log_level)
    os.environ["LOG_LEVEL"] = log_level

    if str(config["config"]["headless"]).lower() == "true":
        os.environ["HEADLESS"] = "true"

    # set agent language and framework as env vars
    os.environ["AGENT_LANGUAGE"] = config["config"]["language"]
    os.environ["AGENT_FRAMEWORK"] = config["config"]["framework"]
    test_dir = get_test_dir()

    # raise an error if required config is missing
    for item in ["url", "prompt"]:
        if config["agent"].get(item) is None:
            err_console.print(
                f"Please set agent.{item} in the config.toml file or pass it as the --{item} CLI option."
            )
            raise typer.Exit()

    # setup the testing environment
    if config["config"]["language"] in ["javascript", "typescript"]:
        node_version = check_for_node()
        npm_version = check_for_npm()

        if not node_version or not npm_version:
            logger.error(
                "Node.js and npm is required to run the agent in JavaScript or TypeScript mode."
            )
            raise typer.Exit()

    # check if playwright is installed
    if config["config"]["framework"] == "playwright":
        playwright_version = check_for_playwright()
        if not playwright_version:
            logger.error("Playwright is required to run the agent in Playwright mode.")
            raise typer.Exit()

    # clean out the test directory
    if str(config["config"]["clean"]).lower() == "true":
        logger.info(f"Deleting files in the {test_dir} directory")
        os.popen(f"rm -rf {test_dir}/*").read()

        if config["config"]["framework"] == "playwright":
            scaffold_playwright(
                test_dir,
                language=config["config"]["language"],
                clean=config["config"]["clean"],
            )
            check_for_playwright_browsers(test_dir)

        if config["config"]["framework"] == "cypress":
            # TODO: Add support for TypeScript
            if config["config"]["language"].lower() == "typescript":
                logger.error("Cypress with TypeScript, is not currently supported.")
                raise typer.Exit()

            scaffold_cypress(test_dir, language=config["config"]["language"])
            check_for_cypress_installation(test_dir)

    # call the agent
    agent(
        prompt=config["agent"]["prompt"],
        url=config["agent"]["url"],
        language=config["config"]["language"],
        framework=config["config"]["framework"],
    )



import typer
import subprocess
from pathlib import Path
import shutil
from agent.logging import logger


def delete_entire_dir(dir: Path | str) -> None:
    """Delete all files and folders in a directory

    :param dir: The directory to delete
    :type dir: Path | str
    """
    shutil.rmtree(Path(dir), ignore_errors=True)


def check_for_node():
    """Checks if node.js in installed on the system"""
    try:
        logger.info("Checking that Node.js is installed.")
        version = subprocess.check_output("node --version", shell=True)
        if version:
            logger.info(f"Found node version: {version.decode().strip()}")

        return version.decode().strip()
    except Exception as e:
        logger.error(
            f"Could not find Node.js installation. Please install it from https://nodejs.org : {e}"
        )
        raise typer.Exit()


def check_for_npm():
    """Checks if npm is installed on the system"""
    try:
        logger.info("Checking that npm is installed.")
        version = subprocess.check_output("npm --version", shell=True)
        if version:
            logger.info(f"Found npm version: {version.decode().strip()}")

        return version.decode().strip()
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False


def check_for_playwright():
    """Checks if playwright is installed on the system"""
    try:
        logger.info("Checking that Playwright is installed.")
        version = subprocess.check_output("npx playwright --version", shell=True)
        if version:
            logger.info(f"Found playwright version: {version.decode().strip()}")

        return version.decode().strip()
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False


def check_for_cypress():
    """Checks if cypress is installed on the system"""
    try:
        logger.info("Checking that Cypress is installed.")
        version = subprocess.check_output("cypress --version", shell=True)
        if version:
            logger.info(f"Found cypress version: {version.decode().strip()}")

        return version.decode().strip()
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False


def scaffold_playwright(
    test_dir: Path | str,
    language: str,
    clean=True,
) -> None:
    """Scaffolds a playwright test environment

    :param test_dir: Path to the test directory
    :type test_dir: Path
    :param language: The language to scaffold the test environment in (python, typescript or javascript)
    :type language: str
    :param clean: Whether to clean out the test directory before scaffolding, defaults to True
    :type clean: bool, optional
    """
    # ensure test_dir is a Path object
    test_dir = Path(test_dir)

    # clean out the test directory
    if clean:
        delete_entire_dir(test_dir)

    # create the test directory
    test_dir.mkdir(exist_ok=True)

    # nothing to do for python
    if language == "python":
        return None

    # run the playwright init command
    logger.info("Running the playwright install command.")
    lang = "Typescript" if language == "typescript" else "js"
    cmd = f"npm init playwright@latest -- --yes --quiet --install-deps --no-examples --browser=chromium --lang={lang} ."

    subprocess.run(
        cmd,
        cwd=test_dir,
        shell=True,
    )

    logger.info("Adding dev dependencies.")
    subprocess.run(
        "npm install uuid --save-dev",
        cwd=test_dir,
        shell=True,
    )


def check_for_playwright_browsers(
    test_dir: Path,
):
    """Checks if playwright browsers are installed on the system"""
    try:
        logger.info("Checking that Playwright browsers are installed.")
        subprocess.run(
            "npx playwright install chromium",
            check=True,
            cwd=test_dir,
            shell=True,
        )
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False
    return True


def check_for_cypress_installation(test_dir: Path):
    """Checks if cypress is installed on the system"""
    try:
        logger.info("Checking that Cypress is installed.")
        subprocess.run(
            "npx --yes cypress install",
            check=True,
            cwd=test_dir,
            shell=True,
        )
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False


def scaffold_cypress(test_dir: Path, language: str, clean=True) -> None:
    """Scaffolds a playwright test environment

    :param test_dir: Path to the test directory
    :type test_dir: Path
    :param language: The language to scaffold the test environment in (python, typescript or javascript)
    :type language: str
    :param clean: Whether to clean out the test directory before scaffolding, defaults to True
    :type clean: bool, optional
    """

    # ensure test_dir is a Path object
    test_dir = Path(test_dir)

    # clean out the test directory
    if clean:
        delete_entire_dir(test_dir)

    # create the test directory
    test_dir.mkdir(exist_ok=True)
    # create e2e folder
    e2e_folder = test_dir / "e2e"
    e2e_folder.mkdir(exist_ok=True)
    # create a support folder
    support_folder = test_dir / "support"
    support_folder.mkdir(exist_ok=True)
    # create a fixtures folder
    fixtures_folder = test_dir / "fixtures"
    fixtures_folder.mkdir(exist_ok=True)

    # nothing to do for python
    if language == "python":
        return None

    # add a gitignore file
    logger.info("Adding a .gitignore file.")
    with open(test_dir / ".gitignore", "w") as f:
        f.write(
            """
node_modules/
.cypress/
            """
        )

    if language == "typescript":
        logger.info("Adding a support/commands.ts file.")
        with open(support_folder / "commands.ts", "w") as f:
            f.write("""""")

        logger.info("Adding a support/e2e.ts file.")
        with open(support_folder / "e2e.ts", "w") as f:
            f.write("""import './commands'

Cypress.on('uncaught:exception', (err, runnable) => {
  // returning false here prevents Cypress from
  // failing the test
  return false
})""")

        # create a package.json file
        with open(test_dir / "package.json", "w") as f:
            f.write("""{
  "name": "agent-tests",
  "version": "1.0.0",
  "type": "module",
  "scripts": {
  },
  "keywords": [],
  "author": "",
  "license": "ISC",
  "devDependencies": {
    "cypress": "latest",
    "uuid": "latest",
    "typescript": "latest"
  }
}""")

        logger.info("Adding a cypress.config.ts file.")
        with open(test_dir / "cypress.config.ts", "w") as f:
            f.write(
                """import { defineConfig } from 'cypress'

export default defineConfig({
  defaultCommandTimeout: 10000,
  e2e: {
    supportFile: "cypress/support/e2e.ts",
  },
})
"""
            )

        logger.info("Adding a tsconfig.json file.")
        with open(test_dir / "tsconfig.json", "w") as f:
            f.write(
                """{
  "compilerOptions": {
    "target": "es5",
    "lib": ["es5", "dom"],
    "types": ["cypress", "node"]
  },
  "include": ["**/*.ts"]
}
"""
            )

    if language == "javascript":
        logger.info("Adding a support/commands.js file.")
        with open(support_folder / "commands.js", "w") as f:
            f.write("""""")

        with open(support_folder / "e2e.js", "w") as f:
            f.write("""import './commands'

Cypress.on('uncaught:exception', (err, runnable) => {
  // returning false here prevents Cypress from
  // failing the test
  return false
})""")

        logger.info("Adding a cypress.config.js file.")
        with open(test_dir / "cypress.config.js", "w") as f:
            f.write(
                """const { defineConfig } = require('cypress')

module.exports = defineConfig({
  defaultCommandTimeout: 10000,
  e2e: {
    supportFile: "cypress/support/e2e.js",
  },
})
"""
            )

        logger.info("Setting up cypress environment.")
        # create a package.json file
        with open(test_dir / "package.json", "w") as f:
            f.write("""{
  "name": "agent-tests",
  "version": "1.0.0",
  "main": "index.js",
  "scripts": {
  },
  "keywords": [],
  "author": "",
  "license": "ISC",
  "devDependencies": {
    "cypress": "latest",
    "uuid": "latest",
    "typescript": "latest"
  }
}""")

    logger.info("Installing npm dependencies.")
    subprocess.run(
        "npm install",
        cwd=test_dir,
        shell=True,
    )



import os
import tomllib
from pathlib import Path
from typing import Literal

# create a folder to store the generated code
SUPPORTED_LANGUAGES = ["python", "typescript", "javascript"]
SUPPORTED_FRAMEWORKS = ["playwright", "cypress"]


def get_test_dir() -> Path:
    x: Literal["cypress", "playwright", "out"] = os.environ.get(
        "AGENT_FRAMEWORK", "playwright"
    )
    test_dir = Path(x) if x else Path("out")
    test_dir.mkdir(exist_ok=True)
    return test_dir


def read_config():
    config_file = Path("config.toml")

    if config_file.exists():
        with open(config_file, "rb") as f:
            config = tomllib.load(f)

    return config



import io
import base64
import os
import pytest
import pytest_playwright
from pathlib import Path
from contextlib import redirect_stdout
from agent.config import get_test_dir
from agent.logging import logger
from agent.video import extract_frames, keep_unique_images
import subprocess


def strip_code_fences(code):
    potential_fences = [
        "```python",
        "```py",
        "```typescript",
        "```ts",
        "```javascript",
        "```js",
        "```",
    ]
    for fence in potential_fences:
        code = code.replace(fence, "")

    return code


def run_pytest_playwright(test_dir: Path) -> str:
    # Create a StringIO buffer to capture the output
    buffer = io.StringIO()
    args = [
        "-v",
        str(test_dir),
        "--screenshot=on",
        "--video=on",
        "--full-page-screenshot",
    ]

    headless = os.environ.get("HEADLESS", False)
    if not headless:
        args.append("--headed")

    # Redirect the stdout to the buffer
    with redirect_stdout(buffer):
        pytest.main(
            args,
            plugins=[pytest_playwright],
        )
    # Get the content of the buffer
    output = buffer.getvalue()
    # Close the buffer
    buffer.close()
    return output


def run_playwright(test_dir: Path) -> str:
    # ensure test_dir exists
    test_dir.mkdir(exist_ok=True)
    # Create a StringIO buffer to capture the output
    buffer = io.StringIO()

    cmd = ["npx", "playwright", "test", "--trace=on", "--reporter=line"]

    # Run the command using subprocess and capture the output
    process = subprocess.Popen(
        cmd, cwd=test_dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )
    stdout, stderr = process.communicate()

    # Write the output to the buffer
    buffer.write(stdout)
    buffer.write(stderr)

    # Get the content of the buffer
    output = buffer.getvalue()
    # Close the buffer
    buffer.close()
    return output


def run_cypress(test_dir: Path, config_file_name: str) -> str:
    # Create a StringIO buffer to capture the output
    buffer = io.StringIO()
    cmd = [
        "npx",
        "cypress",
        "run",
        "--headless",
        "--config-file",
        str(test_dir / config_file_name),
    ]

    # Run the command using subprocess and capture the output
    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )
    stdout, stderr = process.communicate()

    # Write the output to the buffer
    buffer.write(stdout)
    buffer.write(stderr)

    # Get the content of the buffer
    output = buffer.getvalue()
    # Close the buffer
    buffer.close()
    return output


# Function to encode the image
def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def check_for_screenshots():
    screenshots = []
    test_dir = get_test_dir()

    if not test_dir:
        return screenshots

    # unzip any trace.zip folders
    trace_files = list(test_dir.glob("**/*.zip"))
    logger.info(f"Found {len(trace_files)} trace file(s).")
    if trace_files:
        for trace_file in trace_files:
            # extract the contents of the zip file
            try:
                subprocess.run(
                    f"unzip {str(trace_file)} -d {str(trace_file.parent)}",
                    shell=True,
                )
                # find all image files in any subdirectories
                keep_unique_images(test_dir)
            except Exception as e:
                logger.error(f"Error extracting trace file: {e}")

    # check for video files
    video_files = list(test_dir.glob("**/*.webm"))
    logger.info(f"Found {len(video_files)} video(s).")
    if video_files:
        for video_file in video_files:
            # extract frames from the video using moviepy
            extract_frames(video_file)

    logger.info("Checking for screenshots.")

    image_files = list(test_dir.glob("**/*.png"))

    for format in ["jpeg", "jpg", "webp"]:
        image_files.extend(list(test_dir.glob(f"**/*.{format}")))

    logger.info(f"Found {len(image_files)} screenshot(s).")
    if image_files:
        for image_file in image_files:
            # convert to base64 and display in a panel
            image_data = encode_image(image_file)
            screenshots.append(image_data)

    return screenshots



from moviepy.editor import VideoFileClip
import numpy as np
from datetime import timedelta
from pathlib import Path
import math
from PIL import Image, ImageChops
from agent.logging import logger


def format_timedelta(td):
    """Utility function to format timedelta objects in a cool way (e.g 00:00:20.05)
    omitting microseconds and retaining milliseconds"""
    result = str(td)
    try:
        result, ms = result.split(".")
    except ValueError:
        return (result + ".00").replace(":", "-")
    ms = int(ms)
    ms = round(ms / 1e4)
    return f"{result}.{ms:02}".replace(":", "-")


def calculate_rmse(image1_path, image2_path):
    """
    Calculate the Root Mean Square Error (RMSE) between two images.
    """
    with Image.open(image1_path) as img1, Image.open(image2_path) as img2:
        # Convert both images to a consistent format (e.g., RGBA)
        img1 = img1.convert("RGBA")
        img2 = img2.convert("RGBA")

        # Ensure the images have the same size for comparison
        if img1.size != img2.size:
            return float(
                "inf"
            )  # Consider images with different sizes as completely different

        # Calculate the difference between the images
        diff = ImageChops.difference(img1, img2)

        # Calculate the squared difference of each pixel
        h = diff.histogram()
        sum_of_squares = sum(value * ((idx % 256) ** 2) for idx, value in enumerate(h))

        # Calculate the RMSE
        rmse = math.sqrt(sum_of_squares / (img1.size[0] * img1.size[1]))
        return rmse


def keep_unique_images(directory, threshold=5.0):
    """
    Find and keep only unique images in a directory based on pixel content similarity.
    Delete images that are almost identical to others.
    """
    unique_images = []
    image_paths = list(Path(directory).glob("**/*.png"))
    image_paths.extend(list(Path(directory).glob("**/*.jpeg")))
    image_paths.extend(list(Path(directory).glob("**/*.jpg")))
    image_paths.extend(list(Path(directory).glob("**/*.webp")))

    # Loop through all PNG files and compare them pairwise
    while image_paths:
        current_image = image_paths.pop(0)
        unique_images.append(current_image)

        for other_image in image_paths[:]:
            try:
                # Calculate the RMSE between the current image and another
                rmse = calculate_rmse(current_image, other_image)

                # If RMSE is below the threshold, consider them almost identical
                if rmse < threshold:
                    logger.debug(
                        f"Deleting nearly identical image: {other_image} (RMSE: {rmse:.2f})"
                    )
                    other_image.unlink()
                    image_paths.remove(other_image)
            except Exception as e:
                logger.error(f"Error comparing {current_image} and {other_image}: {e}")

    return unique_images


def extract_frames(video_file: Path, frames_per_second: int = 1):
    # load the video clip
    video_clip = VideoFileClip(str(video_file))

    # if the SAVING_FRAMES_PER_SECOND is above video FPS, then set it to FPS (as maximum)
    saving_frames_per_second = min(video_clip.fps, frames_per_second)
    # if SAVING_FRAMES_PER_SECOND is set to 0, step is 1/fps, else 1/SAVING_FRAMES_PER_SECOND
    step = (
        1 / video_clip.fps
        if saving_frames_per_second == 0
        else 1 / saving_frames_per_second
    )

    # iterate over each possible frame
    for current_duration in np.arange(0, video_clip.duration, step):
        # format the file name and save it
        frame_duration_formatted = format_timedelta(timedelta(seconds=current_duration))
        frame_filename = video_file.parent / f"frame{frame_duration_formatted}.png"
        # save the frame with the current duration
        video_clip.save_frame(frame_filename, current_duration)

    keep_unique_images(video_file.parent)



import pytest
from pathlib import Path
from agent.scaffold import scaffold_playwright, delete_entire_dir


@pytest.mark.parametrize("language", ["python", "typescript", "javascript"])
def test_playwright_scaffolding(language: str):
    test_dir = Path(f"test_playwright_{language}")

    scaffold_playwright(
        test_dir=test_dir,
        language=language,
        clean=True,
    )

    files_in_test_dir = list(test_dir.glob("*"))

    # check for the node_modules folder
    node_modules = test_dir / "node_modules"

    if language != "python":
        # check for the node_modules folder
        assert node_modules.exists()

        # check for relevant files
        assert any("package.json" in str(f) for f in files_in_test_dir)
        assert any(".gitignore" in str(f) for f in files_in_test_dir)

    if language == "javascript":
        assert any("playwright.config.js" in str(f) for f in files_in_test_dir)

    if language == "typescript":
        assert any("playwright.config.ts" in str(f) for f in files_in_test_dir)

    # delete the test directory
    delete_entire_dir(test_dir)
